/////////////////////////////////////////////////////////////
//	FILE: graphicsMath.cpp
//  created:  Oct 12 2009 by M. van de Panne, UBC
///////////////////////////////////////////////////////////////

#include "defs.h"
#include "graphicsMath.h"

TMATRIX::TMATRIX() {
	iden();
}

TMATRIX::TMATRIX(VEC3 i, VEC3 j, VEC3 k, VEC3 p) {
	iden();
	for (int r=0; r<3; r++) {
		m[r][0] = i.v[r];
		m[r][1] = j.v[r];
		m[r][2] = k.v[r];
		m[r][3] = p.v[r];
	}
}

void TMATRIX::iden() {
	for (int i=0; i<4; i++) {
		for (int j=0; j<4; j++) {
			m[i][j] = ( i==j? 1 : 0);
		}
	}
}

VEC3 TMATRIX::operator*(const VEC3 a) {
	return VEC3(
		m[0][0]*a.v[0] + m[0][1]*a.v[1] + m[0][2]*a.v[2] + m[0][3],
		m[1][0]*a.v[0] + m[1][1]*a.v[1] + m[1][2]*a.v[2] + m[1][3],
		m[2][0]*a.v[0] + m[2][1]*a.v[1] + m[2][2]*a.v[2] + m[2][3]
	);
}

TMATRIX TMATRIX::operator*(const TMATRIX a) {
	TMATRIX ans;
	for (int r=0; r<4; r++) {
		for (int c=0; c<4; c++) {
			ans.m[r][c] = 0.0;
			for (int n=0; n<4; n++)
				ans.m[r][c] += m[r][n]*a.m[n][c];
		}
	}
	return ans;
}

void TMATRIX::print(const char *name) {
	printf("%s\n",name);
	for (int r=0; r<4; r++) {
		printf("%6.3f %6.3f %6.3f %6.3f\n", m[r][0], m[r][1], m[r][2], m[r][3]);
	}
}

void TMATRIX::rotkth(VEC3 k, float th)
{
    float cth, sth, vth;
    float kx, ky, kz;

	th *= M_PI/180.0f;
    cth = (float) cos(th);
    vth = 1.0f-cth;
    sth = (float) sin(th);
    kx = k.v[0];
    ky = k.v[1];
    kz = k.v[2];
	iden();
    m[0][0] = kx*kx*vth + cth;
    m[0][1] = ky*kx*vth - kz*sth;
    m[0][2] = kz*kx*vth + ky*sth;
    m[1][0] = kx*ky*vth + kz*sth;
    m[1][1] = ky*ky*vth + cth;
    m[1][2] = kz*ky*vth - kx*sth;
    m[2][0] = kx*kz*vth - ky*sth;
    m[2][1] = ky*kz*vth + kx*sth;
    m[2][2] = kz*kz*vth + cth;
}

void TMATRIX::translate(VEC3 t)
{
	iden();
    m[0][3] = t.v[0];
    m[1][3] = t.v[1];
    m[2][3] = t.v[2];
}

VEC3 TMATRIX::getcol(int c)
{
	return VEC3( m[0][c], m[1][c], m[2][c]);
}

void TMATRIX::quat(float q[])
{
	float w,x,y,z;
	float xs,ys,zs,wx,wy,wz,xx,yy,zz,xy,yz,xz;
	
	w = q[0];
	x = q[1];
	y = q[2];
	z = q[3];

	xs = x + x;
	ys = y + y;
	zs = z + z;
	wx = w*xs;
	wy = w*ys;
	wz = w*zs;
	xx = x*xs;
	yy = y*ys;
	zz = z*zs;
	xy = x*ys;
	yz = y*zs;
	xz = x*zs;
	m[0][0] = 1.0f - (yy+zz);
	m[0][1] = xy-wz;
	m[0][2] = xz+wy;
	m[0][3] = 0.0;
	m[1][0] = xy + wz;
	m[1][1] = 1.0f - (xx+zz);
	m[1][2] = yz - wx;
	m[1][3] = 0.0;
	m[2][0] = xz-wy;
	m[2][1] = yz + wx;
	m[2][2] = 1.0f - (xx+yy);
	m[2][3] = 0.0;
	m[3][0] = 0.0;
	m[3][1] = 0.0;
	m[3][2] = 0.0;
	m[3][3] = 1.0f;
}

void TMATRIX::getkth(VEC3 &k, float &th)
{
    float sth, cth;
    float a1,a2,a3;

    cth = 0.5f*(m[0][0] + m[1][1] + m[2][2] - 1.0f);
    a1 = m[2][1] - m[1][2];
    a2 = m[0][2] - m[2][0];
    a3 = m[1][0] - m[0][1];
    sth = 0.5f* (float) sqrt(a1*a1 + a2*a2 + a3*a3);
    if (fabs(sth)<0.0001f) {
      k.v[0] = 1.0f;
      k.v[1] = 0.0f;
      k.v[2] = 0.0f;
      th = 0.0f;
      return;
    }
    th = (float) atan2(sth, cth)*180.0/M_PI;
    k.v[0] = (m[2][1] - m[1][2])/(2.0f*sth);
    k.v[1] = (m[0][2] - m[2][0])/(2.0f*sth);
    k.v[2] = (m[1][0] - m[0][1])/(2.0f*sth);
}

///////////////////////////////////////////////////////////
// graphicsMathTest()
///////////////////////////////////////////////////////////

void graphicsMathTest() {
      // This routine is called from init() at program startup

	printf("\n VEC3 tests\n");
	VEC3 a(1,2,3);     //  create new vectors
	VEC3 b(1,1,1);    
	VEC3 c,e;
	a.print("a");
	b.print("b");
	float len1 = a.len();   // compute vector length, i.e., magnitude
	float len2 = b.len();
	float d = a.dot(b);     // dot product
	printf("len(a)=%6.3f, len(b)=%6.3f, dotprod=%6.3f\n", len1, len2, d );
	c = a + b;       // vector addition
	c.print("c=a+b");
	a += b;
	c.print("a+=b");
	a.normalize();   // vector normalisation
	a.print("a");
	printf("len(a)=%6.3f\n",a.len());
	a = VEC3(1,0,0);
	b = VEC3(0,1,0);
	c = a*b;         // cross product
	printf("cross product of i x j:\n");
	c.print("c");
	e = a + b;
	e.print("e=a+b");
	e = a - b;
	e.print("e=a-b");

	printf("\n TMATRIX tests\n");
	TMATRIX f,g;
	f.print("matrix f");
	  // setup matrix
	  //     [ 0 -1  0 -2 ]
	  // f = [ 1  0  0  1 ]
	  //     [ 0  0  1  0 ]
	  //     [ 0  0  0  1 ]
	f = TMATRIX(           // matrix that does    Trans(-2,1,0)Rot(z,90)
		   VEC3(0,1,0),    // i
		   VEC3(-1,0,0),   // j
		   VEC3(0,0,1),    // k
		   VEC3(-2,1,0) ); // p  i.e., translation
	f.print("matrix f");
	VEC3 point(1,2,0);
	VEC3 pointPrime = f*point;       // p' = M*p   i.e., transform a point
	pointPrime.print("pointPrime");
	g = TMATRIX(           // matrix    Trans(-1,-2,0)Rot(z,-90)
		VEC3(0,-1,0),
		VEC3(1,0,0),
		VEC3(0,0,1),
		VEC3(-1,-2,0) );
	g.print("matrix g");
	TMATRIX h = f*g;              // multiply two transformation matrices
	h.print("matrix h=f*g");
	g.rotkth(VEC3(0,0,1),90);     // g = Rot(z,90)
	g.print("Rot(z,90)");
	g.rotkth(VEC3(0,1,0),90);     // g = Rot(y,90)
	g.print("Rot(y,90)");
	VEC3 k;
	float theta;
	g.getkth(k,theta);            // compute (k,th) from a matrix
	k.print("k");
	printf("theta=%6.3f\n",theta);
}

